# email2toppy.py - setup your Topfield 5100PVR timers using email.
# See http://www.iki.fi/ike/email2toppy
#
# version 0.2 - 2006-06-18 - by Ilkka Salminen (ike@iki.fi)
# This Python script is released to public domain. You are allowed to
# use it as you like. Comments, bugs, etc. are welcomed. 
import sys, imaplib, socket, re, ConfigParser, logging, smtplib, os

DEF_INIFILE = 'email2toppy.cfg'

def readConfig(inifile):
	"""Reads the inifile. 
	Returns settings configuration object"""
	
	config = ConfigParser.SafeConfigParser()
	f = config.read(inifile)
	if(len(f) == 0): 
		return None
	else:
		return config

def setLogParameters(params):
	try:
		loglevel = int(params['loglevel'])
	except KeyError:
		loglevel = 50
	try:
		logfile = params['logfile']
	except KeyError:
		logfile = DEF_INIFILE
	filemode = 'a'
	if('append' in params):
		append = params['append']
		if(append == 'false'):
			filemode = 'w'
	logging.basicConfig(level=loglevel,
			format='%(asctime)s %(levelname)s %(message)s',
			filename=logfile,
            filemode=filemode)

def getImapMail(imapSettings):
	"""Gets messages from IMAP server. 
	Returns new messages with given tag in subject. Returened messages are in list of 2 tuples of (HEADERS, BODY)"""
	try:
		tag = imapSettings["tag"]
		server = imapSettings["server"]
		logging.info("Connecting to IMAP server: "+server)
		mailbox = imaplib.IMAP4(server)
		mailbox.login(imapSettings["user"], imapSettings["password"])
	except KeyError, error:
		logging.critical("Missing IMAP configuration: "+str(error))
		return []
	except imaplib.IMAP4.error, error:
		logging.error("Cannot connect to IMAP server: " + str(error))
		return []
	except socket.error, error:
		logging.error("Network error while connecting to IMAP server: "+str(error))
		return []
		
	imapSearchString = '(UNSEEN SUBJECT "'+tag+'")'
	#imapSearchString = '(SUBJECT "'+tag+'")'
	logging.debug("Using IMAP search string: "+imapSearchString)
	messageList = []
	try:
		mailbox.select()
		ok, data = mailbox.search(None, imapSearchString)
		if(ok != 'OK'): 
			logging.error("IMAP server error: "+ok)
			return []
		else: 
			logging.info("Connected to IMAP server")
		msgNums = data[0].split();
		logging.debug("Fetching messages: "+ str(msgNums))
		for msg in msgNums:
			ok, data = mailbox.fetch(msg, '(RFC822.HEADER RFC822.TEXT)')
			messageList.append((data[0][1], data[1][1]))
	except imaplib.IMAP4.error, error:
		logging.error("IMAP error: "+str(error))
	
	logging.info("Closing IMAP connection")
	mailbox.close()
	mailbox.logout()
	return messageList		

def parseMessage(message, paramSyntax, realPIN):
	"""Parses an email with valid PIN. paramSyntax contains regular expressions for allowed parameter values.
	Returns tuple (reply address if log was requested, list of remote commands). So far only one 
	command per email is allowed."""

	if(message == ()):
		logging.error("parseMessage: empty email")
		return ([],[])
	headers = message[0]
	body = message[1]
	replyAddress = ''
	keys = paramSyntax.keys()
	regExpKeys = '|'.join(keys)
	timerItemRegExp = r'^('+regExpKeys+r')=(.+)$'
	regItems = re.compile(timerItemRegExp, re.MULTILINE)
	regExpReply = r'^(From|ReplyTo): (.+@.+)$'
	regReply = re.compile(regExpReply, re.MULTILINE)
	commands = []
	items = {}
	for key, value in regItems.findall(body):
		items[key] = value.strip()
	errorItems = checkParameters(paramSyntax, items)
	givenPin = ''
	if('PIN' in items):
		givenPin = items['PIN']
	if(givenPin == realPin):
		if('log' in items and items['log'].lower() == 'yes'):
			foundReplyAddresses = dict(regReply.findall(headers))
			logging.debug("Found these replyAddresses: "+str(foundReplyAddresses))
			if('ReplyTo' in foundReplyAddresses):
				replyAddress = foundReplyAddresses['ReplyTo']
				logging.debug("Using replyto address: "+replyAddress)
			elif('From' in foundReplyAddresses):
				replyAddress = foundReplyAddresses['From']
				logging.debug("Using from address: "+replyAddress)
			else:
				logging.info("No reply address defined")
			start = replyAddress.find('<') # if reply contains < > characters, take email address from between
			if(start >= 0): 
				end = replyAddress.rfind('>')
				if(end > start):
					replyAddress = replyAddress[start+1:end]
					logging.info("Stripped real name from reply address. Using: "+replyAddress)
		if(len(errorItems) > 0):
			logging.error("Incorrect parameter: "+str(errorItems))
		else:	
			try:
				cmdString = ''
				cmdString += "Command."+items["command"]+":"+items["channel"]+","+items["date"]+","+items["start"]+","+items["duration"]+","+"TV"+","+items["repeat"]+","+items["type"]+";"+items["name"]
				commands.append(cmdString)
			except KeyError, error:
				logging.info("Incomplete command, missing parameter "+str(error))
	else: # no valid PIN
		logging.info('Invalid PIN received: '+givenPin)
		logging.debug('Correct PIN is: '+realPin)

	return (replyAddress, commands)
		
def checkParameters(paramSyntax, params):
	"""Checks the format of the emailed parameters. paramSyntax contains regular expressions for 
	allowed parameter values. Params is a dictionaty of parameters   Returns a list of errors."""
	errors = []
	for param, value in params.iteritems():
		if(not re.match(paramSyntax[param], value)):
			errors.append(param)
	return errors		
		
def makeCommandFile(cmdFileName, commands):
	"""Creates commandfile to be sent to Toppy. 
	Returns True if no errors, False otherwise. """
	logging.debug("Creating command data")
	
	try:
		f = file(cmdFileName, 'w')
		for cmd in commands:
			f.write(cmd+'\n')
		f.close()
	except IOError, error:
		logging.critical("Error creating commandfile: "+str(error))
		return False
	return True
			
def sendFileToToppy(transferSettings):
	"""Sends JAG's EPG remote command to Toppy."""
	logging.debug("Sending commands to Toppy")
	try:
		command = transferSettings['writecmd']
	except KeyError, error:
		logging.critical("Missing transfer configuration: "+str(error))
		return False
	logging.info("Issuing command "+command)
	retVal = os.system(command)
	if(retVal != 0):
		logging.error("Sending commandfile failed")
		return False
	
	return True
	
	
def getResultsFromToppy(transferSettings):
	"""Reads timerlist from Toppy
	Return string containing the timerlist"""
	logging.debug("Reading timerlist from Toppy")
	try:
		readCommand = transferSettings['readcmd']
		deleteCommand = transferSettings['deletecmd']
	except KeyError, error:
		logging.critical("Missing transfer configuration: "+str(error))
		return False
		
	logging.info("Issuing command "+readCommand)
	retVal = os.system(readCommand)
	if(retVal != 0):
		logging.info("Reading results failed. Maybe no results or no connection to Toppy")
		return False
	elif(deleteCommand.strip() != ''):
		logging.debug("Issuing command: "+deleteCommand)
		retVal = os.system(deleteCommand)
		if(retVal != 0):
			logging.error("Could not delete returnfile")
			return False
	
	return True

	
def emailData(replysTo, resultFileName, logFileName, smtpSettings):
	"""Emails data to address in replysTo and in replyto address in config file"""
	toAddr = ""
	try:
		fromAddr = smtpSettings['from']
		configToAddr = smtpSettings['replyto']
		if(configToAddr not in replysTo):
			replysTo.append(configToAddr)
		toAddr = ','.join(replysTo)
		headers = "From: "+fromAddr
		headers += "\r\n" + "To: "+toAddr
		headers += "\r\nSubject: "+smtpSettings['subject']
		server = smtpSettings['server']
	except KeyError, error:
		logging.critical("Missing SMTP configuration: "+str(error))
		return
	body = ""
	# read the results from file
	try:
		f = open(resultFileName)
		body = f.read()
		f.close()
	except IOError, error:
		logging.info("Can't read resultfile: "+str(error))

	# if log is requested, read the log
	if(len(replysTo) > 0):
		try:
			f = open(logFileName)
			body += "\r\n============================ LOG FILE ================================\r\n"
			body += f.read()
			f.close()
		except IOError, error:
			logging.info("Can't read logfile: "+str(error))
			return
		
		
	msg = headers+"\r\n\r\n"+body+"\r\n"
	logging.info("Emailing results to "+str(replysTo))
	try:
		server = smtplib.SMTP(server)
		server.sendmail(fromAddr, replysTo, msg)
		logging.info("Email sent ok")
	except smtplib.SMTPException, error:
		logging.critical("Error sending reply: "+str(error))
	
	server.quit()


	
if __name__ == "__main__":
	if(len(sys.argv) > 1):
		inifile = sys.argv[1]
	else:
		inifile = DEF_INIFILE
	config = readConfig(inifile)
	if(config == None):
		logging.critical("No config file")
		sys.exit(1)
	setLogParameters(dict(config.items('log')))
	logging.info("Configuration file used: "+inifile)

	paramSyntax = {
		'PIN'		: r'\w+', 				# PIN can be any non empty string 
		'channel'	: r'\d{4}',				# channel must be 4 digits
		'command'	: r'AddTimer',			# at this time AddTimer is only supported command
		'date'		: r'\d{4}-\d{2}-\d{2}',	# date in format YYYY-MM-DD
		'start'		: r'\d{2}:\d{2}',		# start time in format hh:mm
		'duration'	: r'\d{2}:\d{2}',		# duration in hh:mm
		'type'		: r'Rec|-P-',			# Rec for recording, -P- for turning the device on
		'repeat'	: r'OT|ED|WE|WL|WD',	# OT= one time, ED=every day, etc.
		'name'		: r'\w+',				# at least one character
		'log'		: r'yes|no'				# yes or no
		}

	imapSettings = dict(config.items('imap'))
	mails = []
	allCommands = []
	replysTo = []
	realPin = config.get('general', 'pin')
	if(realPin == ''):
		logging.info("No PIN defined - no emails")
	else:
		if(len(imapSettings) == 0):
			logging.info("No IMAP settings defined - can only return data from Toppy")
		else:
			mails = getImapMail(imapSettings)
	ok = False
	logRequested = False
	for msg in mails:
		replyAddress, commands = parseMessage(msg, paramSyntax, realPin)
		if(len(replyAddress) > 0): # at least one reply address is here  i.e. log is asked
			logRequested = True
		allCommands.extend(commands)
		if(replyAddress not in replysTo):
			replysTo.append(replyAddress)
	if(len(allCommands) > 0):
		allCommands.append('Command.CreateTimerList')
	
	transferSettings = dict(config.items('transfer'))
	if(len(allCommands) > 0):
		ok = False
		try:
			ok= makeCommandFile(transferSettings['localdir']+transferSettings['commandfile'], allCommands)
		except KeyError, error:
			logging.error("Undefined configuration setting: "+str(error))
		if(ok):
			sendFileToToppy(transferSettings)
	else:
		logging.info("No messages")
		
	ok = getResultsFromToppy(transferSettings)
	if(ok or logRequested):
		try:
			emailData(replysTo, transferSettings['localdir']+transferSettings['returnfile'], config.get('log', 'logfile'), dict(config.items('smtp')))
		except KeyError, error:
			logging.error("Missing configuration parameter: "+str(error))
		
	